/**************************************************************************/
/***** Programs: LMT SITE EVALUATION RADIOMETER LINEAR DATA REDUCTION *****/
/***** NAME:    linear_red.c			  		      *****/
/***** AUTHOR:	David Hiriart	MODIFIED: Douglas McGonagle	      *****/
/***** DATE:	9211.00						      *****/
/**************************************************************************/

/*** REVISION HISTORY
	Nov. 1992 (D.H.)
	Feb. 1994 (D.H.)
	Sep. 1994 (D.H.)
	Oct. 1994 (DMcG)
	Jan. 1995 (DMcG)
	Jul. 1995 (DMcG)
	Oct. 1995 (D.H.) 
		- Correction for zener non-linearities at ADC input. 
		- Correction for the reference voltage for ADC converter
		- It doesn't exit at Numerical Recipes Errors
		- It flags numerical recipes errors (1) Warning errors (2) or
		  both (3)
		- Corrects offset from temperature display
		- Corrects the standar deviation
	Nov. 1995 (DMcG)
		- Added a third command line argument for the sr library path.
		- Modified the set_lookup_table file name argument to include
		  the sr/lib path.
		- Removed TEMP_OFFSET.
		- Modified the A/D calibration and zener effect interpolation.
		- Vcc and VGunn voltages are first divided by a factor of 2 and
		  3, respectively, before making calibration corrections and 
		  then scaled back.
	Jan, 1996 (DMcG)
		- Added a test on the sky voltages, to make sure they always
		  increase in value with increasing air mass (4), and V_sky is 
		  always less than V_Ref (8).
	          		      ERROR CODES
		  ====================================================
			BIT	DESCRIPTION
		  ----------------------------------------------------
			 0	WARNING message in data file
			 1      NR ERROR occured during fit
			 2      V_sky does not increase with Air Mass
			 3      V_sky > V_ref
			 4      Fit tau < 0.0
			 5      Alternate Vref_ADC used.
		  ====================================================
		- Modified coefficients for determing sigmas for old data sets.
			v_sig:   0.01 -> 0.003
		        ref_sig: 0.01 -> 0.002
		  New coefficients based on SR5 data.
		- Will not attempt fit if error_flag >= 8
		- To prevent formatting problems:
			If chi2f > 999.0, then set = 999.0
			If sbfit >  99.0, then set =  99.0
		        if afit  > 9.9, then set =   9.9
		- Added a line to the file header wich includes the name of the
		  inut data file.
		- Changed the forced settings of chi2f  and tau_sig after
		  an NR ERROR from  0 to 999, 99 and 9.999, respectively.
		- Added code to check for DIAGNOSTIC messages (lines starting
		  with the character %), and clear them out.
		- Some SR units have transitory or persistant low voltage 
		  measurements. This appears to be due to a fault in the ref.
	          voltage source. The ref. voltage source appears to jump from
	          Vref to ~Vcc; though, some evidence suggests there may also
		  be intermediate values at times. To correct for this, there
		  are now two sets of vref_ADC vectors, (1) and (2). (1) is 
		  the normal Vref value and (2) is set to a value which makes
		  the Vcc and Vgunn voltages match "normal" data. If we don't 
		  have transitory data, we default to setting (2) to 5V.
		- Error in determing the Vcc and Vgunn voltages for old data.
		  Voltages were being divided by factors of 2 and 3, 
	 	  respectively. This is only neccessary for SDF data.
	June 1996  (DMcG)
		 - Defined MIROFF to be the offset in the mirror postion. 
		   Radiometer beam position on sky (air mass) is now calculated
		   using ANGOFF and MIROFF. Spill-over correction s_fac[i] uses
		   xx[i] which is calculated only with MIROFF (note: change in
		   box angle dosn't effect the spill over, only an offset in 
		   the mirror position).
***/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include <setjmp.h>
#include <string.h>

static jmp_buf mark_place;
FILE *fp,*fpo, *fito;
char time_stamp[80],inp_file[256];
float *x,*y,*sig;

/*** BRING IN TE NUMERICAL RECIPES SUBROUTINES FOR THE LINEAR FIT ***/
#include "Mnrutil.c"
#include "gammq.c"
#include "gcf.c"
#include "gser.c"
#include "gammln.c"
#include "fit.c"

#define IGNORE  1
#define SI      1
#define NO      0
#define DEG_RAD 3.141592654/180.0
#define SQR(x)  ((x)*(x))
#define REDANG  8               /* Number of angles to determine tau */
#define GAINOFF 1.0		/* Gain applied to Vsky */
#define ANGOFF  0.0  		/* Box Angle offset (degrees) */ 
#define MIROFF  2.5             /* Mirror Angle Offset (degrees) */
#define DAYOF	0.0		/* Day Offset */
#define HROF	0.0		/* Hour Offset */
#define V_OFFSET 6.0		/* Offset at the ADC input in mV */
#define TABLE_SIZE 1000		/*Size of the table correction  */


#include "set_table.c"		/* Function to set up lookup table */
#include "linterpol.c"		/* Function for linear interpolation */

				/* Get 'n' arguments from char string 'line'
				   and return them in 'args' array */
int getargs(line,n,args)
int n;
char *line;
float args[];
{
        int i,j,k;
        char numb[15];
        j=3;
        for(i=0;i<n;i++){
                while(isdigit(line[j])==0) j++;
                k=0;
                while((isdigit(line[j])!=0) || (line[j]=='.')){
                        numb[k]=line[j];
                        j++; k++;
                }
                numb[k]='\0';
                args[i]=atof(numb);
        }
	return 0;
}
				/* Bemporad's polynomial approximation
				   for the air mass */

int salida()
{
    fclose(fp);
    fclose(fpo);
    fclose(fito);
    free_vector(sig,1,REDANG);
    free_vector(y,1,REDANG);
    free_vector(x,1,REDANG);
    exit(1);
 }

int main(argc, argv)
int  argc;
char *argv[];
{
    int i,j, SDF=0, WARN, table_size,unit,error_flag,retval;
    char string[180],ad_str[80];
    float hour,min,sec,day,time,t1,t2,t3,t4,v1,v2;
    float args[6],ang[8],ma[8],v[8],ref[8],tet,t,taum,num;
    float afit,bfit,safit,sbfit,chi2f,qfit;
    float sigma1,sigma2;
    float v_sig[8], ref_sig[8],Ref_Sig[8];
    float resid,sum,s_fac, xx[9];
    float *vin,*vout,factor;
    /* Sin Uso: */
    /*
    float Sigma,error, drefm,refm, refm_s,q2fit,timin;
    float a2fit,b2fit,sa2fit,sb2fit,chi2f2,offset;
    char erof[8],infile[10];
    int flag=0,perm=SI;
    */
    /*** A/D Calibration ***/
    static char *arch[]={
    "ad6.dat","ad6.dat","ad6.dat","ad6.dat","ad6.dat","ad6.dat","ad6.dat"};

    /*** A/D reference voltage at Pin 28, measured relative to TP1 (mV). ***/
    /*** NOTE: Vref for unit 5 is not known. 4500 is an average 	 ***/
    /* float vref1_ADC[7]={4500.0,4510.0,4470.0,4490.0,4510.0,4500.0,4521.0}; */
    /*** OK: Unit 5 is measured at IAR, it is now 4416.6 mV              ***/
    float vref1_ADC[7]={4500.0,4510.0,4470.0,4490.0,4510.0,4416.6,4521.0};

    /*** PROBLEMS WITH SR3, VREF MAY HAVE CHANGED TO 5V 9601.27          ***/
    /***  + SR3 SETTING vref2_ADC TO 5V APPEARS TO FIX LOW VOLTAGE       ***/
    /***    MEASUREMENTS. NO TRANSITION DATA TO FIT vref2_ADC VALUE.     ***/
    /***    THE SENCOND Vref WAS SET BY MATCHING Vgunn TO THE SHIPPING   ***/
    /***    CHECK SHEET.						 ***/
    /***  + SR5 REAL FLAKEY, DETERMIND vref2_ADC BY FITTING TO REAL      ***/
    /***    su950818 data.  						 ***/
    float vref2_ADC[7]={4500.0,4510.0,4470.0,5055.0,4510.0,4955.0,4521.0};

    /*** A/D analog ground voltage at Pin 27, measureds relative to  ***/
    /*** TP1 (mV) with the stepper motor off.			    ***/
    float agrd_ADC[7] = {6.0,6.0,6.0,6.0,6.0,6.0,6.0};

    x   = vector(1,REDANG);
    y   = vector(1,REDANG);
    sig = vector(1,REDANG);

    if(argc == 1) {
       fprintf(stderr,
      "   Syntax: %s <input_data.file> [<ouput.file>]\n\n",argv[0]);
       exit(1);
    }
     
    if((fp=fopen(argv[1],"r"))==NULL){
        fprintf(stderr,"Error opening file %s data input.\n\n",argv[1]);
        exit(1);
    }

    strcpy(inp_file,argv[1]);

    do{
   	if(fgets(string,100,fp) == NULL)
		salida();
    }while(string[0] != '*');

    if(isdigit(string[1])){
    	SDF = 1;			/* Check if Standard Data Format */
	unit = string[1] - 0x30;         /* Gets Radiometer Unit */
    }
    if(string[1] == '*'){
        SDF = 0;			/* Old data format */
	unit = 0;
    }
    rewind(fp);

    /* Set lookup table to correct for zener non_linearities at ADC input*/

    sprintf(ad_str,"%s",arch[unit]);
    vin = vector(1,TABLE_SIZE);
    vout= vector(1,TABLE_SIZE);
    set_lookup_table(ad_str,vin,vout,&table_size); 

    if(argc < 3) {
       if((fpo=fopen("prelimi.dat","w+"))==NULL){
	  fprintf(stderr,"Error opening file prelimi.dat.\n\n");
          exit(1);
       }
    } else {
       if((fpo=fopen(argv[2],"w"))==NULL){
	  fprintf(stderr,"Error opening file %s for output.\n\n", argv[2]);
          exit(1);
       }
    }

    if((fito=fopen("fit.dat","w"))==NULL){
        fprintf(stderr,"Error opening file fit.dat for output.\n\n");
        exit(1);
    }
	                  /* Data Input */

   printf("# INPUT FILE: %s\n", argv[1]);
   printf(
     "# LINEAR: UNIT=#%1d REDANG=%2d GAINOFF=%6.3f ANGOF=%6.3f MIROFF=%6.3f\n",
     unit,REDANG, GAINOFF, ANGOFF, MIROFF);
   printf("#         DAYOF = %8.3f  HROF = %7.3f\n", DAYOF, HROF);
   printf("#         vref1_ADC = %6.1f  agrd_ADC = %4.1f\n",
     vref1_ADC[unit],agrd_ADC[unit]);   
   printf("#         ad_str: %s\n", ad_str);
   printf(
"# TIME    TAU    CHI2   Tamb   Tins   TAU_S  W  RESID     5V       Vg      b\n");

   fprintf(fpo,"# INPUT FILE: %s\n", argv[1]);
   fprintf(fpo,
     "# LINEAR: UNIT=#%1d REDANG=%2d GAINOFF=%6.3f ANGOF=%6.3f MIROFF=%6.3f\n",
     unit,REDANG, GAINOFF, ANGOFF, MIROFF);
   fprintf(fpo,"#            DAYOF = %8.3f  HROF = %7.3f\n", DAYOF, HROF);
   fprintf(fpo,"#            vref1_ADC = %6.1f  agrd_ADC = %4.1f\n",
     vref1_ADC[unit],agrd_ADC[unit]);   
   fprintf(fpo,"#            ad_str: %s\n", ad_str);
   fprintf(fpo,
"# TIME    TAU    CHI2   Tamb   Tins   TAU_S  W  RESID     5V       Vg      b\n");

    error_flag = 0;

    do{
	/*** Test if an NR error occured ***/
	retval = setjmp(mark_place);
	if(retval == 1){
		taum = 0.0;
		chi2f = 999.0;
		sbfit = 99.0;
		resid = 9.999;
		afit = 9.999;
		error_flag |= 2;
   printf("%7.3f %6.3f %7.3f %6.2f %6.2f %6.3f %2d %6.3f %8.2f %8.2f %6.3f\n",
           t, taum, chi2f, t2, t4, sbfit, error_flag,resid, v1, v2, afit);
   fprintf(fpo,
         "%7.3f %6.3f %7.3f %6.2f %6.2f %6.3f %2d %6.3f %8.2f %8.2f %6.3f\n",
           t, taum, chi2f, t2, t4, sbfit,error_flag,resid, v1, v2, afit);
	}

	if (error_flag != 0) 
    fprintf(stderr,"(%s) WARNING: %3d, AT: %s",inp_file,error_flag,time_stamp);

	error_flag = WARN = 0;

        do{
                if(fgets(string,100,fp)==NULL) salida();
	}while(string[0]!='*');

					/*Time header */
	strcpy(time_stamp,string);
        getargs(string,5,args);
	hour=args[0]+HROF;
        min =args[1];
        sec =args[2];
	day =args[3]+DAYOF;
	time=args[4];
        t = day+(hour*3600.0+min*60.0+sec+time)/(3600.0*24); 
      
					/* Temps. and Voltages*/
        if(fgets(string,160,fp)==NULL) salida();
        if (SDF) {
	   getargs(string,8,args);
	} else {
	   getargs(string,6,args);   
	}

        /*** Check for Vref goinbt to Vcc, adjust factor accordingly ***/
        /* Set correction factor for the reference voltage of the UNIT ADC */
        factor = ((vref1_ADC[unit]-agrd_ADC[unit])/4500.0); /* NORMAL */

	if(SDF){
	   //fprintf(stderr,"v1  = %f\n",args[4]);
   	   v1  = ((args[4]*0.5)+agrd_ADC[unit])*factor;
	   //fprintf(stderr,"v1a = %f\n",v1);
	   v1 += linterpol(v1,vin,vout,table_size);
	   //fprintf(stderr,"v1b = %f\n",v1);
	   v1 *= 2.0;
	   //fprintf(stderr,"v1c = %f\n",v1);
	}else{
   	   v1  = ((args[4])+agrd_ADC[unit])*factor;
	   v1 += linterpol(v1,vin,vout,table_size);
	   v1 *= 2.0;	
	}

        /* If v1 is low, use the backup vref2_ADC value */
	if (v1 < 4600.0) {
	   error_flag |= 32;
           factor = ((vref2_ADC[unit]-agrd_ADC[unit])/4500.0); /* Vref=Vcc */
	}

        /*** Corrected voltages measured by AD converter.            ***/
	/*** voltages are first corrected for A/D ref. voltage error ***/
	/*** and are then corrected for the non-linear zener error   ***/
   	t1  = (args[0]+agrd_ADC[unit])*factor;
       	t1 += linterpol(t1,vin,vout,table_size);
        t2  = (args[1]+agrd_ADC[unit])*factor;
	t2 += linterpol(t2,vin,vout,table_size);
   	t3  = (args[2]+agrd_ADC[unit])*factor;
	t3 += linterpol(t3,vin,vout,table_size);
        t4  = (args[3]+agrd_ADC[unit])*factor;
	t4 += linterpol(t4,vin,vout,table_size);

	/*** NOTE: Sigma's are calculated from the difference of two  ***/
	/*** voltages which removes the offset due to agrd_ADC	      ***/
	if(SDF) {
   	   v1  = ((args[4]*0.5)+agrd_ADC[unit])*factor;
	   v1 += linterpol(v1,vin,vout,table_size);
	   v1 *= 2.0;
 	   sigma1  = args[5]*factor;
	   sigma1 += linterpol(sigma1,vin,vout,table_size);
     	   v2  = ((args[6]/3.0)+agrd_ADC[unit])*factor;
	   v2 += linterpol(v2,vin,vout,table_size);
	   v2 *= 3.0; 
 	   sigma2  = args[7]*factor;
	   sigma2 += linterpol(sigma2,vin,vout,table_size);
	} else {
   	   v1  = ((args[4])+agrd_ADC[unit])*factor;
	   v1 += linterpol(v1,vin,vout,table_size);
	   v1 *= 2.0;
     	   v2  = ((args[5])+agrd_ADC[unit])*factor;
	   v2 += linterpol(v2,vin,vout,table_size);
	   v2 *= 3.0;
	   sigma1 = v1*0.0002;
           sigma2 = v2*0.0002;
	}

				    	/* Scan */
	for(i=0;i<8;i++){

	      /*** CHECK FOR DIAGNOSTIC LINES STARTING WITH % ***/
	      do{
                 if(fgets(string,100,fp)==NULL) salida();
	      }while(string[0]=='%');

                if(isalpha(string[0])||isalpha(string[3])){ 
		   if(fgets(string,100,fp)==NULL) salida();
		   WARN = 1;
		}
		if(isalpha(string[0])||isalpha(string[3])){ 
		   if(fgets(string,100,fp)==NULL) salida();
		   WARN = 1;
		   }
		if (SDF) {
		   getargs(string,6,args);
		} else {
		   getargs(string,4,args);
		}
                ang[i]=args[0];
                ma[i] =args[1];

     		v[i]  = (args[2]+agrd_ADC[unit])*factor;
     		v[i] += linterpol(v[i],vin,vout,table_size);
	        if (SDF) {
  		   v_sig[i]  = args[3]*factor;
		   v_sig[i] += linterpol(v_sig[i],vin,vout,table_size);
   		   ref[i]    = (args[4]+agrd_ADC[unit])*factor;
		   ref[i]   += linterpol(ref[i],vin,vout,table_size); 
		   ref_sig[i]  = args[5]*factor;
		   ref_sig[i] += linterpol(ref_sig[i],vin,vout,table_size);
		} else {
    		   ref[i]  = (args[3]+agrd_ADC[unit])*factor;
	           ref[i] += linterpol(ref[i],vin,vout,table_size);
    		   v_sig[i]   = v[i]*0.0003;
    		   ref_sig[i] = ref[i]*0.0002;		   
		}
	}

	if(WARN == 1) error_flag |= 1;

	/*** It corrects sigmas for reference ***/
	Ref_Sig[0] = 1.41421*ref_sig[0];
	for(i=1; i<8; i++)
             Ref_Sig[i] = sqrt(fabs(4.0*SQR(ref_sig[i])-2.0*SQR(Ref_Sig[0])));
	ref_sig[0] = 0.5*sqrt(2.0*SQR(Ref_Sig[0]));
	for(i=1; i<8; i++)
		ref_sig[i] = 0.5*sqrt(SQR(Ref_Sig[i])+SQR(Ref_Sig[i-1]));

	/*** Test v[j], make sure value increases with j ***/
	for(j=1; j<REDANG; j++) 
	   if(v[j-1] >= v[j]) error_flag |= 4;
	
	/* Data transformation */
        for(j=0;j<REDANG;j++){
                tet=ma[j];
		tet=acos(1.0/tet);
	        /* Radiometer Beam Position on Sky (air mass) */
                x[j+1]=1.0/cos(tet+(ANGOFF*DEG_RAD)+(2.0*MIROFF*DEG_RAD));
	        /* Radiometer Mirror Position (air mass) */
	        xx[j+1]=1.0/cos(tet+(2.0*MIROFF*DEG_RAD));
                num = ref[j] - GAINOFF*v[j];

	        /*** Test if num is greater than zero ***/
                if(num <=0.0) {
		   num=1.0;
                   error_flag |= 8;
		}
		/*** SPILLOVER CORRECTION FACTOR:   ***/
		/*** Determined from SR6, July 1995 ***/
		s_fac = 1.0 - (0.0208*xx[j+1] - 0.0164);
                y[j+1]=log(num/s_fac);

		sig[j+1]=sqrt(SQR(v_sig[j])+SQR(ref_sig[j]))/num;
		fprintf(fito," %f %f %f %f\n", x[j+1], y[j+1], sig[j+1],s_fac);
	}

     /*** Fit only if error_flag BIT-3 is not true ***/
     if ((error_flag & 8)==0) { 		/* Fit a straight line */
	fit(x,y,REDANG,sig,SDF,&afit,&bfit,&safit,&sbfit,&chi2f,&qfit);
        fprintf(fito," %f %f %f\n", afit, bfit, chi2f);
	taum = -bfit;

	/*** Calculate residuals ***/
 	for(sum = 0.0, j=1;j<=REDANG;j++){
	   resid= y[j]-(bfit*x[j]+afit);
	   sum += SQR(resid);
	}
	resid = sqrt((double)sum);

   }else{
      taum = 0;
      chi2f = 999.0;
      sbfit = 99.0;
      resid = 9.999;
      afit = 0;
   }

   if (taum < 0.0) error_flag |= 16;

   if (chi2f > 999.0) chi2f = 999.0;
   if (resid > 99.0)  resid = 99.0;
   if (sbfit > 9.999) sbfit = 9.999;
   if(taum>0.0){
   printf("%7.3f %6.3f %7.3f %6.2f %6.2f %6.3f %2d %6.3f %8.2f %8.2f %6.3f\n",
           t, taum, chi2f,t2,t4, sbfit, error_flag,resid, v1, v2, afit);
   fprintf(fpo,
         "%7.3f %6.3f %7.3f %6.2f %6.2f %6.3f %2d %6.3f %8.2f %8.2f %6.3f\n",
           t, taum, chi2f,t2, t4, sbfit, error_flag,resid, v1, v2, afit);
   }
}while(1);

    fclose(fp);
    fclose(fpo);
    fclose(fito);
    free_vector(vin,1,TABLE_SIZE);
    free_vector(vout,1,TABLE_SIZE);
    free_vector(sig,1,REDANG);
    free_vector(y,1,REDANG);
    free_vector(x,1,REDANG);
    return 0;
}

